import logging
from typing import AsyncGenerator, Iterator, Union
import os
import io

from pydub import AudioSegment

from achatbot.common.interface import ITts
from achatbot.common.session import Session
from achatbot.common.types import GTTS_SYNTHESIS_FILE, GTTSArgs, RECORDS_DIR
from .base import BaseTTS


class GTTS(BaseTTS, ITts):
    TAG = "tts_g"

    @classmethod
    def get_args(cls, **kwargs) -> dict:
        return {**GTTSArgs().__dict__, **kwargs}

    def __init__(self, **args) -> None:
        self.args = GTTSArgs(**args)
        self.file_path = os.path.join(RECORDS_DIR, GTTS_SYNTHESIS_FILE)

    async def _inference(self, session: Session, text: str) -> AsyncGenerator[bytes, None]:
        from gtts import gTTS
        with io.BytesIO() as f:
            tts = gTTS(
                text=text,
                lang=self.args.language,
                tld=self.args.tld,
                slow=self.args.slow)
            tts.write_to_fp(f)
            f.seek(0)

            audio: AudioSegment = AudioSegment.from_mp3(f)
            if self.args.speed_increase != 1.0:
                audio = audio.speedup(
                    playback_speed=self.args.speed_increase,
                    chunk_size=self.args.chunk_size,
                    crossfade=self.args.crossfade_lenght)
            audio_resampled = audio.set_frame_rate(
                22050).set_channels(
                1).set_sample_width(2)  # 16bit sample_width 16/8=2
            audio_data = audio_resampled.raw_data
            yield audio_data

    def get_voices(self):
        from gtts.lang import tts_langs
        voices = []
        languages = tts_langs()
        tlds = ['com', 'com.au', 'co.uk', 'us', 'ca', 'co.in', 'ie', 'co.za']

        for lang in languages.keys():
            for tld in tlds:
                voices.append(GTTSArgs(language=lang, tld=tld))

        return voices



import os
import logging

from achatbot.common import interface
from achatbot.common.factory import EngineClass, EngineFactory
from achatbot.common.types import AudioPlayerArgs

from dotenv import load_dotenv
load_dotenv(override=True)


class PlayerEnvInit():
    @staticmethod
    def initPlayerEngine() -> interface.IPlayer | EngineClass:
        from . import stream_player
        # player
        tag = os.getenv('PLAYER_TAG', "stream_player")
        kwargs = AudioPlayerArgs(
            is_immediate_stop=bool(os.getenv('IS_IMMEDIATE_STOP', "")),
        ).__dict__
        engine = EngineFactory.get_engine_by_tag(EngineClass, tag, **kwargs)
        logging.info(f"initPlayerEngine: {tag},  {engine}")
        return engine



# Copyright (c) 2021 Mobvoi Inc. (authors: Binbin Zhang)
#               2024 Alibaba Inc (authors: Xiang Lyu)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import torchaudio


def read_lists(list_file):
    lists = []
    with open(list_file, 'r', encoding='utf8') as fin:
        for line in fin:
            lists.append(line.strip())
    return lists

def read_json_lists(list_file):
    lists = read_lists(list_file)
    results = {}
    for fn in lists:
        with open(fn, 'r', encoding='utf8') as fin:
            results.update(json.load(fin))
    return results

def load_wav(wav, target_sr):
    speech, sample_rate = torchaudio.load(wav)
    speech = speech.mean(dim=0, keepdim=True)
    if sample_rate != target_sr:
        assert sample_rate > target_sr, 'wav sample rate {} must be greater than {}'.format(sample_rate, target_sr)
        speech = torchaudio.transforms.Resample(orig_freq=sample_rate, new_freq=target_sr)(speech)
    return speech



from dataclasses import dataclass
from typing import Any, List


from apipeline.frames.data_frames import Frame, DataFrame, TextFrame, ImageRawFrame


@dataclass
class URLImageRawFrame(ImageRawFrame):
    """An image with an associated URL. Will be shown by the transport if the
    transport's camera is enabled.

    """
    url: str | None

    def __str__(self):
        return f"{self.name}(url: {self.url}, size: {self.size}, format: {self.format})"


@dataclass
class VisionImageRawFrame(ImageRawFrame):
    """An image with an associated text to ask for a description of it. Will be
    shown by the transport if the transport's camera is enabled.

    """
    text: str | None

    def __str__(self):
        return f"{self.name}(text: {self.text}, size: {self.size}, format: {self.format})"


@dataclass
class UserImageRawFrame(ImageRawFrame):
    """An image associated to a user. Will be shown by the transport if the
    transport's camera is enabled.

    """
    user_id: str

    def __str__(self):
        return f"{self.name}(user: {self.user_id}, size: {self.size}, format: {self.format})"


@dataclass
class SpriteFrame(Frame):
    """An animated sprite. Will be shown by the transport if the transport's
    camera is enabled. Will play at the framerate specified in the transport's
    `fps` constructor parameter.

    """
    images: List[ImageRawFrame]

    def __str__(self):
        return f"{self.name}(size: {len(self.images)})"


@dataclass
class TranscriptionFrame(TextFrame):
    """A text frame with transcription-specific data. Will be placed in the
    transport's receive queue when a participant speaks.

    """
    user_id: str
    timestamp: str

    def __str__(self):
        return f"{self.name}(user: {self.user_id}, text: {self.text}, timestamp: {self.timestamp})"


@dataclass
class InterimTranscriptionFrame(TextFrame):
    """A text frame with interim transcription-specific data. Will be placed in
    the transport's receive queue when a participant speaks."""
    user_id: str
    timestamp: str

    def __str__(self):
        return f"{self.name}(user: {self.user_id}, text: {self.text}, timestamp: {self.timestamp})"


@dataclass
class LLMMessagesFrame(DataFrame):
    """A frame containing a list of LLM messages. Used to signal that an LLM
    service should run a chat completion and emit an LLMStartFrames, TextFrames
    and an LLMEndFrame. Note that the messages property on this class is
    mutable, and will be be updated by various ResponseAggregator frame
    processors.

    """
    messages: List[dict]


@dataclass
class TransportMessageFrame(DataFrame):
    message: Any

    def __str__(self):
        return f"{self.name}(message: {self.message})"


@dataclass
class LLMMessagesAppendFrame(DataFrame):
    """A frame containing a list of LLM messages that neeed to be added to the
    current context.

    """
    messages: List[dict]


@dataclass
class LLMMessagesUpdateFrame(DataFrame):
    """A frame containing a list of new LLM messages. These messages will
    replace the current context LLM messages and should generate a new
    LLMMessagesFrame.

    """
    messages: List[dict]


@dataclass
class TTSSpeakFrame(DataFrame):
    """A frame that contains a text that should be spoken by the TTS in the
    pipeline (if any).

    """
    text: str


@dataclass
class DailyTransportMessageFrame(TransportMessageFrame):
    participant_id: str | None = None



import os
import time
import asyncio
from common.interface import IBuffering
from common.session import Session
from common.types import SilenceAtEndOfChunkArgs


class SilenceAtEndOfChunk(IBuffering):
    def __init__(self, args: SilenceAtEndOfChunkArgs) -> None:
        self.chunk_length_seconds = os.environ.get(
            'BUFFERING_CHUNK_LENGTH_SECONDS')
        if not self.chunk_length_seconds:
            self.chunk_length_seconds = args.chunk_length_seconds
        self.chunk_length_seconds = float(self.chunk_length_seconds)

        self.chunk_offset_seconds = os.environ.get(
            'BUFFERING_CHUNK_OFFSET_SECONDS')
        if not self.chunk_offset_seconds:
            self.chunk_offset_seconds = args.chunk_offset_seconds
        self.chunk_offset_seconds = float(self.chunk_offset_seconds)

        self.processing_flag = False
        self.buffer = bytearray()
        self.scratch_buffer = bytearray()

    def insert(self, audio_data):
        self.buffer.extend(audio_data)

    def clear(self):
        self.buffer.clear()
        self.scratch_buffer.clear()

    def process_audio(self, session: Session):
        chunk_length_in_bytes = self.chunk_length_seconds * \
            session.ctx.sampling_rate * session.ctx.sample_width
        if len(self.buffer) > chunk_length_in_bytes:
            if self.processing_flag:
                exit("Error in realtime processing: tried processing a new chunk while the previous one was still being processed")

            self.scratch_buffer += session.buffer
            self.buffer.clear()
            self.processing_flag = True
            # Schedule the processing in a separate task on asyncio event loop
            asyncio.create_task(self.process_audio_async(session))

    def is_voice_active(self, session) -> bool:
        if "vad_res" not in session.ctx.state \
                or len(session.ctx.state["vad_res"]) == 0:
            return False
        last_segment_should_end_before = ((len(self.scratch_buffer) / (
            session.ctx.sampling_rate * session.ctx.sample_width))
            - self.chunk_offset_seconds)
        return session.ctx.state["vad_res"][-1]['end'] < last_segment_should_end_before

    async def process_audio_async(self, session: Session):
        if session.ctx.vad is None \
                or session.ctx.asr is None:
            self.processing_flag = False
            return

        start = time.time()
        vad_res = await session.ctx.vad.detect(session)
        session.ctx.state["vad_res"] = vad_res

        if self.is_voice_active(session):
            transcription = await session.ctx.asr.transcribe(session)
            if transcription['text'] != '':
                end = time.time()
                transcription['processing_time'] = end - start
            self.scratch_buffer.clear()

        self.processing_flag = False




from achatbot.common.factory import EngineClass


class BaseLLM(EngineClass):
    def model_name(self):
        if hasattr(self.args, "model_name"):
            return self.args.model_name
        return ""



import logging


from achatbot.common.utils.audio_utils import convertSampleRateTo16khz
from achatbot.common.session import Session
from achatbot.common.interface import IDetector
from achatbot.common.types import (
    RATE, CHUNK
)
from achatbot.common.factory import EngineClass


class BaseVAD(EngineClass, IDetector):
    async def detect(self, session: Session):
        pass

    def get_sample_info(self):
        return RATE, CHUNK

    def set_audio_data(self, audio_data):
        if isinstance(audio_data, (bytes, bytearray)):
            self.audio_buffer = audio_data
            if hasattr(self.args, "sample_rate") is False:
                return
            if self.args.sample_rate != 16000:
                audio_data_16 = convertSampleRateTo16khz(audio_data, self.args.sample_rate)
                logging.debug(
                    f"rate {self.args.sample_rate} convertSampleRateTo16khz len(audio_data):{len(audio_data)} -> len(audio_data_16):{len(audio_data_16)}")
                self.audio_buffer = audio_data_16

    def close(self):
        pass



import base64
import json
import os
import logging
from typing import List


try:
    from openai import AsyncOpenAI, AsyncStream, DefaultAsyncHttpxClient
    from openai.types.chat import (
        ChatCompletionChunk,
        ChatCompletionFunctionMessageParam,
        ChatCompletionMessageParam,
        ChatCompletionToolParam
    )
except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error(
        "In order to use OpenAI, you need to `pip install openai`. Also, set `OPENAI_API_KEY` environment variable.")
    raise Exception(f"Missing module: {e}")
import httpx
from apipeline.frames.data_frames import TextFrame, Frame
from apipeline.pipeline.pipeline import FrameDirection

from achatbot.processors.aggregators.openai_llm_context import OpenAILLMContext, OpenAILLMContextFrame
from achatbot.processors.llm.base import LLMProcessor
from achatbot.types.frames.control_frames import LLMFullResponseEndFrame, LLMFullResponseStartFrame, LLMModelUpdateFrame
from achatbot.types.frames.data_frames import LLMMessagesFrame, VisionImageRawFrame


class OpenAIUnhandledFunctionException(Exception):
    pass


class BaseOpenAILLMProcessor(LLMProcessor):
    """This is the base for all processors that use the AsyncOpenAI client.

    This processor consumes OpenAILLMContextFrame frames, which contain a reference
    to an OpenAILLMContext frame. The OpenAILLMContext object defines the context
    sent to the LLM for a completion. This includes user, assistant and system messages
    as well as tool choices and the tool, which is used if requesting function
    calls from the LLM.
    """

    def __init__(self, *, model: str, api_key="", base_url="", **kwargs):
        super().__init__(**kwargs)
        api_key = os.environ.get("OPENAI_API_KEY", api_key)
        self._model: str = model
        self._client = AsyncOpenAI(
            api_key=api_key,
            base_url=base_url,
            http_client=DefaultAsyncHttpxClient(
                limits=httpx.Limits(
                    max_keepalive_connections=100,
                    max_connections=1000,
                    keepalive_expiry=None,
                )
            )
        )

    def can_generate_metrics(self) -> bool:
        return True

    async def get_chat_completions(
            self,
            context: OpenAILLMContext,
            messages: List[ChatCompletionMessageParam]) -> AsyncStream[ChatCompletionChunk]:
        chunks = await self._client.chat.completions.create(
            model=self._model,
            stream=True,
            messages=messages,
            tools=context.tools,
            tool_choice=context.tool_choice,
        )
        return chunks

    async def _stream_chat_completions(
            self, context: OpenAILLMContext) -> AsyncStream[ChatCompletionChunk]:
        logging.debug(f"Generating chat: {context.get_messages_json()}")

        messages: List[ChatCompletionMessageParam] = context.get_messages()

        # base64 encode any images
        for message in messages:
            if message.get("mime_type") == "image/jpeg":
                encoded_image = base64.b64encode(message["data"].getvalue()).decode("utf-8")
                text = message["content"]
                message["content"] = [
                    {"type": "text", "text": text},
                    {"type": "image_url", "image_url": {"url": f"data:image/jpeg;base64,{encoded_image}"}}
                ]
                del message["data"]
                del message["mime_type"]

        chunks = await self.get_chat_completions(context, messages)

        return chunks

    async def _process_context(self, context: OpenAILLMContext):
        function_name = ""
        arguments = ""
        tool_call_id = ""

        await self.start_ttfb_metrics()

        chunk_stream: AsyncStream[ChatCompletionChunk] = (
            await self._stream_chat_completions(context)
        )

        async for chunk in chunk_stream:
            if len(chunk.choices) == 0:
                continue

            await self.stop_ttfb_metrics()

            if chunk.choices[0].delta.tool_calls:
                # We're streaming the LLM response to enable the fastest response times.
                # For text, we just yield each chunk as we receive it and count on consumers
                # to do whatever coalescing they need (eg. to pass full sentences to TTS)
                #
                # If the LLM is a function call, we'll do some coalescing here.
                # If the response contains a function name, we'll yield a frame to tell consumers
                # that they can start preparing to call the function with that name.
                # We accumulate all the arguments for the rest of the streamed response, then when
                # the response is done, we package up all the arguments and the function name and
                # yield a frame containing the function name and the arguments.

                tool_call = chunk.choices[0].delta.tool_calls[0]
                if tool_call.function and tool_call.function.name:
                    function_name += tool_call.function.name
                    tool_call_id = tool_call.id
                    await self.call_start_function(function_name)
                if tool_call.function and tool_call.function.arguments:
                    # Keep iterating through the response to collect all the argument fragments
                    arguments += tool_call.function.arguments
            elif chunk.choices[0].delta.content:
                await self.push_frame(TextFrame(chunk.choices[0].delta.content))

        # if we got a function name and arguments, check to see if it's a function with
        # a registered handler. If so, run the registered callback, save the result to
        # the context, and re-prompt to get a chat answer. If we don't have a registered
        # handler, raise an exception.
        if function_name and arguments:
            if self.has_function(function_name):
                await self._handle_function_call(context, tool_call_id, function_name, arguments)
            else:
                raise OpenAIUnhandledFunctionException(
                    f"The LLM tried to call a function named '{function_name}', but there isn't a callback registered for that function.")

    async def _handle_function_call(
            self,
            context: OpenAILLMContext,
            tool_call_id: str,
            function_name: str,
            arguments: str
    ):
        arguments = json.loads(arguments)
        result = await self.call_function(function_name, arguments)
        arguments = json.dumps(arguments)
        if isinstance(result, (str, dict)):
            # Handle it in "full magic mode"
            tool_call = ChatCompletionFunctionMessageParam({
                "role": "assistant",
                "tool_calls": [
                    {
                        "id": tool_call_id,
                        "function": {
                            "arguments": arguments,
                            "name": function_name
                        },
                        "type": "function"
                    }
                ]
            })
            context.add_message(tool_call)
            if isinstance(result, dict):
                result = json.dumps(result)
            tool_result = ChatCompletionToolParam({
                "tool_call_id": tool_call_id,
                "role": "tool",
                "content": result
            })
            context.add_message(tool_result)
            # re-prompt to get a human answer
            await self._process_context(context)
        elif isinstance(result, list):
            # reduced magic
            for msg in result:
                context.add_message(msg)
            await self._process_context(context)
        elif isinstance(result, type(None)):
            pass
        else:
            raise TypeError(f"Unknown return type from function callback: {type(result)}")

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        context = None
        if isinstance(frame, OpenAILLMContextFrame):
            context: OpenAILLMContext = frame.context
        elif isinstance(frame, LLMMessagesFrame):
            context = OpenAILLMContext.from_messages(frame.messages)
        elif isinstance(frame, VisionImageRawFrame):
            context = OpenAILLMContext.from_image_frame(frame)
        elif isinstance(frame, LLMModelUpdateFrame):
            logging.debug(f"Switching LLM model to: [{frame.model}]")
            self._model = frame.model
        else:
            await self.push_frame(frame, direction)

        if context:
            await self.push_frame(LLMFullResponseStartFrame())
            await self.start_processing_metrics()
            await self._process_context(context)
            await self.stop_processing_metrics()
            await self.push_frame(LLMFullResponseEndFrame())


class OpenAILLMProcessor(BaseOpenAILLMProcessor):
    TAG = "openai_llm_processor"

    def __init__(self, model: str = "gpt-4o", **kwargs):
        super().__init__(model=model, **kwargs)



import logging
import re
from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from typing import AsyncGenerator, Generator

import numpy as np

from achatbot.common.types import PYAUDIO_PAINT16
from achatbot.common.factory import EngineClass
from achatbot.common.session import Session
from achatbot.common.utils import task


class TTSVoice:
    def __init__(self, name, id):
        self.name = name
        self.id = id

    def __repr__(self):
        return f"<TTSVoice(name={self.name} id={self.id})>"


class BaseTTS(EngineClass):
    def synthesize_sync(self, session: Session) -> Generator[bytes, None, None]:
        is_stream = self.args.tts_stream if hasattr(self.args, "tts_stream") else False
        logging.debug(f"is_stream:{is_stream}")
        if is_stream is False:
            buff = bytearray()
            scratch_buff = bytearray()
            stream_info = self.get_stream_info()
            chunk_length_seconds = 3
            if hasattr(self.args, "chunk_length_seconds"):
                chunk_length_seconds = self.args.chunk_length_seconds
            chunk_length_in_bytes = chunk_length_seconds * \
                stream_info["rate"] * stream_info["sample_width"]

        queue: Queue = Queue()
        with ThreadPoolExecutor() as executor:
            executor.submit(task.fetch_async_items, queue, self.synthesize, session)

            while True:
                item = queue.get()
                if is_stream is True:
                    if item is None:
                        break
                    yield item
                else:
                    if item is None:
                        if len(buff) > 0:
                            yield bytes(buff)
                        buff.clear()
                        scratch_buff.clear()
                        break
                    if len(buff) > chunk_length_in_bytes:
                        logging.debug(
                            f"len(buff) {len(buff)} > chunk_length_in_bytes {chunk_length_in_bytes} to yield")
                        scratch_buff = buff
                        yield bytes(scratch_buff)
                        buff.clear()
                    else:
                        buff.extend(item)

    async def synthesize(self, session: Session) -> AsyncGenerator[bytes, None]:
        if "tts_text_iter" in session.ctx.state:
            for text in session.ctx.state["tts_text_iter"]:
                text: str = self.filter_special_chars(text)
                if len(text.strip()) == 0:
                    continue
                async for chunk in self._inference(session, text):
                    yield chunk
                silence_chunk = self._get_end_silence_chunk(session, text)
                if silence_chunk:
                    yield silence_chunk
        elif "tts_text" in session.ctx.state:
            text = session.ctx.state["tts_text"]
            text = self.filter_special_chars(text)
            if len(text.strip()) > 0:
                async for chunk in self._inference(session, text):
                    yield chunk
                silence_chunk = self._get_end_silence_chunk(session, text)
                if silence_chunk:
                    yield silence_chunk

    async def _inference(self, session: Session, text: str) -> AsyncGenerator[bytes, None]:
        raise NotImplementedError(
            "The _inference method must be implemented by the derived subclass.")

    def _get_end_silence_chunk(self, session: Session, text: str) -> bytes:
        # Send silent audio
        info = self.get_stream_info()
        sample_rate = info["rate"]
        np_dtype = info['np_dtype']

        end_sentence_delimeters = ".。!！?？…。¡¿"
        mid_sentence_delimeters = ";；:：,，\n（()）【[]】「{}」-“”„\"—/|《》"

        silence_duration = 0.3
        if text[-1] in end_sentence_delimeters:
            if hasattr(self.args, "tts_sentence_silence_duration"):
                silence_duration = self.args.tts_sentence_silence_duration
        elif text[-1] in mid_sentence_delimeters:
            if hasattr(self.args, "tts_comma_silence_duration"):
                silence_duration = self.args.tts_comma_silence_duration
        else:
            if hasattr(self.args, "tts_default_silence_duration"):
                silence_duration = self.args.tts_default_silence_duration

        silent_samples = int(sample_rate * silence_duration)
        silent_chunk = np.zeros(silent_samples, dtype=np_dtype)
        res = silent_chunk.tobytes()
        logging.debug(f"add silent_chunk {silent_chunk.shape},tobytes len {len(res)}")
        return res

    def get_stream_info(self) -> dict:
        return {
            "format": PYAUDIO_PAINT16,
            "channels": 1,
            "rate": 22050,
            "sample_width": 2,
            "np_dtype": np.int16,
        }

    def filter_special_chars(self, text: str) -> str:
        # @TODO: use nlp stream process sentence
        special_chars = ".。,，;；!！?？」>》}\\]】\"”'‘)~"
        return self._filter_special_chars(special_chars, text)

    def remove_trailing_special_chars(self, special_chars: str, s: str) -> str:
        pattern = rf'(.*?)([{special_chars}])[{special_chars}]*$'
        return re.sub(pattern, r'\1\2', s)

    def _filter_special_chars(self, special_chars: str, text: str) -> str:
        pattern = rf'[{special_chars}]'
        match = re.search(pattern, text)
        res = text
        if match:
            first_special_index = match.start()
            res = text[:first_special_index + 1] + \
                re.sub(rf'[{special_chars}]+$', '', text[first_special_index + 1:])
            if len(res) == 1:  # have a special char, return empty
                res = ""
        res = res.strip('\n')
        logging.debug(f"text:{text} --filter--> res:{res} with match:{match}")
        return res

    def set_voice(self, voice: str):
        pass



import logging
from typing import AsyncGenerator, Literal

try:
    from openai import AsyncOpenAI, BadRequestError
except ModuleNotFoundError as e:
    logging.error(
        "In order to use OpenAI, you need to `pip install openai`. Also, set `OPENAI_API_KEY` environment variable.")
    raise Exception(f"Missing module: {e}")
from apipeline.frames.sys_frames import ErrorFrame
from apipeline.frames.data_frames import Frame, AudioRawFrame

from achatbot.processors.speech.tts.base import TTSProcessorBase


class OpenAITTSProcessor(TTSProcessorBase):
    """This processor uses the OpenAI TTS API to generate audio from text.
    The returned audio is PCM encoded at 24kHz. When using the DailyTransport, set the sample rate in the DailyParams accordingly:
    ```
    DailyParams(
        audio_out_enabled=True,
        audio_out_sample_rate=24_000,
    )
    ```
    """

    def __init__(
            self,
            *,
            api_key: str | None = None,
            voice: Literal["alloy", "echo", "fable", "onyx", "nova", "shimmer"] = "alloy",
            model: Literal["tts-1", "tts-1-hd"] = "tts-1",
            **kwargs):
        super().__init__(**kwargs)

        self._voice = voice
        self._model = model

        self._client = AsyncOpenAI(api_key=api_key)

    def can_generate_metrics(self) -> bool:
        return True

    async def set_voice(self, voice: str):
        logging.debug(f"Switching TTS voice to: [{voice}]")
        self._voice = voice

    async def run_tts(self, text: str) -> AsyncGenerator[Frame, None]:
        logging.debug(f"Generating TTS: [{text}]")

        try:
            await self.start_ttfb_metrics()

            async with self._client.audio.speech.with_streaming_response.create(
                    input=text,
                    model=self._model,
                    voice=self._voice,
                    response_format="pcm",
            ) as r:
                if r.status_code != 200:
                    error = await r.text()
                    logging.error(
                        f"{self} error getting audio (status: {r.status_code}, error: {error})")
                    yield ErrorFrame(f"Error getting audio (status: {r.status_code}, error: {error})")
                    self._tts_done_event.set()
                    return
                async for chunk in r.iter_bytes(8192):
                    if len(chunk) > 0:
                        await self.stop_ttfb_metrics()
                        frame = AudioRawFrame(audio=chunk, sample_rate=24_000, num_channels=1)
                        yield frame
                self._tts_done_event.set()
        except BadRequestError as e:
            logging.exception(f"{self} error generating TTS: {e}")
            self._tts_done_event.set()



from . import dummy_bot


def import_bots(bot_name: str = "DummyBot"):
    """ import package to register """
    if "Dummy" in bot_name:
        from . import dummy_bot
        return True
    if "DailyRTVI" in bot_name:
        from . import daily_rtvi_bot
        return True
    if "DailyAsrRTVI" in bot_name:
        from . import daily_asr_rtvi_bot
        return True
    if "DailyLangchainRAG" in bot_name:
        from .rag import daily_langchain_rag_bot
        return True
    if "Daily" in bot_name:
        from . import daily_rtvi_bot
        from . import daily_asr_rtvi_bot
        from .rag import daily_langchain_rag_bot
        return True

    return False



import traceback
import logging
import json
import time
import os
import uuid

import grpc

from achatbot.modules.speech.player import PlayerEnvInit
from achatbot.modules.speech.tts import TTSEnvInit
from achatbot.common.grpc.idl.tts_pb2 import (
    LoadModelRequest, LoadModelResponse,
    SynthesizeRequest, SynthesizeResponse,
)
from achatbot.common.grpc.idl.tts_pb2_grpc import TTSStub
from achatbot.common.grpc.interceptors.authentication_client import add_authentication
from achatbot.common.logger import Logger
from achatbot.common.types import SessionCtx
from achatbot.common.session import Session

Logger.init(logging.DEBUG, app_name="chat-bot-tts-client", is_file=True, is_console=True)


def load_model(channel):
    tag = os.getenv("TTS_TAG", "tts_edge")
    is_reload = bool(os.getenv("IS_RELOAD", None))
    kwargs = TTSEnvInit.map_config_func[tag]()
    tts_stub = TTSStub(channel)
    request = LoadModelRequest(tts_tag=tag, is_reload=is_reload, json_kwargs=json.dumps(kwargs))
    logging.debug(request)
    response = tts_stub.LoadModel(request)
    logging.debug(response)


def synthesize_us(channel):
    tts_stub = TTSStub(channel)
    request_data = SynthesizeRequest(tts_text="你好，我是机器人")
    response_iterator = tts_stub.SynthesizeUS(request_data)
    for response in response_iterator:
        yield response.tts_audio


logging.basicConfig(level=logging.DEBUG)

"""
TTS_TAG=tts_edge python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_g IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_coqui IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_chat IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_cosy_voice IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
"""
if __name__ == "__main__":
    try:
        client_id = str(uuid.uuid4())
        session = Session(**SessionCtx(client_id).__dict__)
        # todo: up to the rpc gateway to auth
        token = "oligei-tts"
        authentication = add_authentication('authorization', token)
        port = os.getenv('PORT', "50052")
        channel = grpc.insecure_channel(f'localhost:{port}')
        channel = grpc.intercept_channel(channel, authentication)

        load_model(channel)
        tts_audio_iter = synthesize_us(channel)

        player = PlayerEnvInit.initPlayerEngine()
        player.start(session)
        for tts_audio in tts_audio_iter:
            logging.debug(f"play tts_chunk len:{len(tts_audio)}")
            session.ctx.state["tts_chunk"] = tts_audio
            player.play_audio(session)
        player.stop(session)
    except grpc.RpcError as e:
        logging.error(f"grpc.RpcError: {e}")
    except Exception as e:
        tb_str = traceback.format_exc()
        logging.error(f"Exception: {e}; traceback: {tb_str}")
    finally:
        channel and channel.close()
        player and player.close()



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
# Copyright FunASR (https://github.com/alibaba-damo-academy/FunASR). All Rights Reserved.
#  MIT License  (https://opensource.org/licenses/MIT)

import sys
from funasr import AutoModel

model_dir = "iic/SenseVoiceSmall"
input_file = (
    "https://isv-data.oss-cn-hangzhou.aliyuncs.com/ics/MaaS/ASR/test_audio/asr_example_zh.wav"
)

model = AutoModel(
    model=model_dir,
    trust_remote_code=True,
)

res = model.generate(
    input=input_file,
    cache={},
    language="auto", # "zn", "en", "yue", "ja", "ko", "nospeech"
    use_itn=False,
)

print(res)



# coding=utf-8

import os
import librosa
import base64
import io
import gradio as gr
import re

import numpy as np
import torch
import torchaudio


from funasr import AutoModel

model = "iic/SenseVoiceSmall"
model = AutoModel(model=model,
				  vad_model="iic/speech_fsmn_vad_zh-cn-16k-common-pytorch",
				  vad_kwargs={"max_single_segment_time": 30000},
				  trust_remote_code=True,
				  )

import re

emo_dict = {
	"<|HAPPY|>": "😊",
	"<|SAD|>": "😔",
	"<|ANGRY|>": "😡",
	"<|NEUTRAL|>": "",
	"<|FEARFUL|>": "😰",
	"<|DISGUSTED|>": "🤢",
	"<|SURPRISED|>": "😮",
}

event_dict = {
	"<|BGM|>": "🎼",
	"<|Speech|>": "",
	"<|Applause|>": "👏",
	"<|Laughter|>": "😀",
	"<|Cry|>": "😭",
	"<|Sneeze|>": "🤧",
	"<|Breath|>": "",
	"<|Cough|>": "🤧",
}

emoji_dict = {
	"<|nospeech|><|Event_UNK|>": "❓",
	"<|zh|>": "",
	"<|en|>": "",
	"<|yue|>": "",
	"<|ja|>": "",
	"<|ko|>": "",
	"<|nospeech|>": "",
	"<|HAPPY|>": "😊",
	"<|SAD|>": "😔",
	"<|ANGRY|>": "😡",
	"<|NEUTRAL|>": "",
	"<|BGM|>": "🎼",
	"<|Speech|>": "",
	"<|Applause|>": "👏",
	"<|Laughter|>": "😀",
	"<|FEARFUL|>": "😰",
	"<|DISGUSTED|>": "🤢",
	"<|SURPRISED|>": "😮",
	"<|Cry|>": "😭",
	"<|EMO_UNKNOWN|>": "",
	"<|Sneeze|>": "🤧",
	"<|Breath|>": "",
	"<|Cough|>": "😷",
	"<|Sing|>": "",
	"<|Speech_Noise|>": "",
	"<|withitn|>": "",
	"<|woitn|>": "",
	"<|GBG|>": "",
	"<|Event_UNK|>": "",
}

lang_dict =  {
    "<|zh|>": "<|lang|>",
    "<|en|>": "<|lang|>",
    "<|yue|>": "<|lang|>",
    "<|ja|>": "<|lang|>",
    "<|ko|>": "<|lang|>",
    "<|nospeech|>": "<|lang|>",
}

emo_set = {"😊", "😔", "😡", "😰", "🤢", "😮"}
event_set = {"🎼", "👏", "😀", "😭", "🤧", "😷",}

def format_str(s):
	for sptk in emoji_dict:
		s = s.replace(sptk, emoji_dict[sptk])
	return s


def format_str_v2(s):
	sptk_dict = {}
	for sptk in emoji_dict:
		sptk_dict[sptk] = s.count(sptk)
		s = s.replace(sptk, "")
	emo = "<|NEUTRAL|>"
	for e in emo_dict:
		if sptk_dict[e] > sptk_dict[emo]:
			emo = e
	for e in event_dict:
		if sptk_dict[e] > 0:
			s = event_dict[e] + s
	s = s + emo_dict[emo]

	for emoji in emo_set.union(event_set):
		s = s.replace(" " + emoji, emoji)
		s = s.replace(emoji + " ", emoji)
	return s.strip()

def format_str_v3(s):
	def get_emo(s):
		return s[-1] if s[-1] in emo_set else None
	def get_event(s):
		return s[0] if s[0] in event_set else None

	s = s.replace("<|nospeech|><|Event_UNK|>", "❓")
	for lang in lang_dict:
		s = s.replace(lang, "<|lang|>")
	s_list = [format_str_v2(s_i).strip(" ") for s_i in s.split("<|lang|>")]
	new_s = " " + s_list[0]
	cur_ent_event = get_event(new_s)
	for i in range(1, len(s_list)):
		if len(s_list[i]) == 0:
			continue
		if get_event(s_list[i]) == cur_ent_event and get_event(s_list[i]) != None:
			s_list[i] = s_list[i][1:]
		#else:
		cur_ent_event = get_event(s_list[i])
		if get_emo(s_list[i]) != None and get_emo(s_list[i]) == get_emo(new_s):
			new_s = new_s[:-1]
		new_s += s_list[i].strip().lstrip()
	new_s = new_s.replace("The.", " ")
	return new_s.strip()

def model_inference(input_wav, language, fs=16000):
	# task_abbr = {"Speech Recognition": "ASR", "Rich Text Transcription": ("ASR", "AED", "SER")}
	language_abbr = {"auto": "auto", "zh": "zh", "en": "en", "yue": "yue", "ja": "ja", "ko": "ko",
					 "nospeech": "nospeech"}
	
	# task = "Speech Recognition" if task is None else task
	language = "auto" if len(language) < 1 else language
	selected_language = language_abbr[language]
	# selected_task = task_abbr.get(task)
	
	# print(f"input_wav: {type(input_wav)}, {input_wav[1].shape}, {input_wav}")
	
	if isinstance(input_wav, tuple):
		fs, input_wav = input_wav
		input_wav = input_wav.astype(np.float32) / np.iinfo(np.int16).max
		if len(input_wav.shape) > 1:
			input_wav = input_wav.mean(-1)
		if fs != 16000:
			print(f"audio_fs: {fs}")
			resampler = torchaudio.transforms.Resample(fs, 16000)
			input_wav_t = torch.from_numpy(input_wav).to(torch.float32)
			input_wav = resampler(input_wav_t[None, :])[0, :].numpy()
	
	
	merge_vad = True #False if selected_task == "ASR" else True
	print(f"language: {language}, merge_vad: {merge_vad}")
	text = model.generate(input=input_wav,
						  cache={},
						  language=language,
						  use_itn=True,
						  batch_size_s=0, merge_vad=merge_vad)
	
	print(text)
	text = text[0]["text"]
	text = format_str_v3(text)
	
	print(text)
	
	return text


audio_examples = [
    ["example/zh.mp3", "zh"],
    ["example/yue.mp3", "yue"],
    ["example/en.mp3", "en"],
    ["example/ja.mp3", "ja"],
    ["example/ko.mp3", "ko"],
    ["example/emo_1.wav", "auto"],
    ["example/emo_2.wav", "auto"],
    ["example/emo_3.wav", "auto"],
    #["example/emo_4.wav", "auto"],
    #["example/event_1.wav", "auto"],
    #["example/event_2.wav", "auto"],
    #["example/event_3.wav", "auto"],
    ["example/rich_1.wav", "auto"],
    ["example/rich_2.wav", "auto"],
    #["example/rich_3.wav", "auto"],
    ["example/longwav_1.wav", "auto"],
    ["example/longwav_2.wav", "auto"],
    ["example/longwav_3.wav", "auto"],
    #["example/longwav_4.wav", "auto"],
]




html_content = """
<div>
    <h2 style="font-size: 22px;margin-left: 0px;">Voice Understanding Model: SenseVoice-Small</h2>
    <p style="font-size: 18px;margin-left: 20px;">SenseVoice-Small is an encoder-only speech foundation model designed for rapid voice understanding. It encompasses a variety of features including automatic speech recognition (ASR), spoken language identification (LID), speech emotion recognition (SER), and acoustic event detection (AED). SenseVoice-Small supports multilingual recognition for Chinese, English, Cantonese, Japanese, and Korean. Additionally, it offers exceptionally low inference latency, performing 7 times faster than Whisper-small and 17 times faster than Whisper-large.</p>
    <h2 style="font-size: 22px;margin-left: 0px;">Usage</h2> <p style="font-size: 18px;margin-left: 20px;">Upload an audio file or input through a microphone, then select the task and language. the audio is transcribed into corresponding text along with associated emotions (😊 happy, 😡 angry/exicting, 😔 sad) and types of sound events (😀 laughter, 🎼 music, 👏 applause, 🤧 cough&sneeze, 😭 cry). The event labels are placed in the front of the text and the emotion are in the back of the text.</p>
	<p style="font-size: 18px;margin-left: 20px;">Recommended audio input duration is below 30 seconds. For audio longer than 30 seconds, local deployment is recommended.</p>
	<h2 style="font-size: 22px;margin-left: 0px;">Repo</h2>
	<p style="font-size: 18px;margin-left: 20px;"><a href="https://github.com/FunAudioLLM/SenseVoice" target="_blank">SenseVoice</a>: multilingual speech understanding model</p>
	<p style="font-size: 18px;margin-left: 20px;"><a href="https://github.com/modelscope/FunASR" target="_blank">FunASR</a>: fundamental speech recognition toolkit</p>
	<p style="font-size: 18px;margin-left: 20px;"><a href="https://github.com/modelscope/CosyVoice" target="_blank">CosyVoice</a>: high-quality multilingual TTS model</p>
</div>
"""


def launch():
	with gr.Blocks(theme=gr.themes.Soft()) as demo:
		# gr.Markdown(description)
		gr.HTML(html_content)
		with gr.Row():
			with gr.Column():
				audio_inputs = gr.Audio(label="Upload audio or use the microphone")
				
				with gr.Accordion("Configuration"):
					# task_inputs = gr.Radio(choices=["Speech Recognition", "Rich Text Transcription"],
					# 					   value="Speech Recognition", label="Task")
					language_inputs = gr.Dropdown(choices=["auto", "zh", "en", "yue", "ja", "ko", "nospeech"],
												  value="auto",
												  label="Language")
				fn_button = gr.Button("Start", variant="primary")
				text_outputs = gr.Textbox(label="Results")
			gr.Examples(examples=audio_examples, inputs=[audio_inputs, language_inputs], examples_per_page=20)
		
		fn_button.click(model_inference, inputs=[audio_inputs, language_inputs], outputs=text_outputs)
		# with gr.Accordion("More examples"):
		# 	gr.HTML(centered_table_html)
	demo.launch()


if __name__ == "__main__":
	# iface.launch()
	launch()





import collections
import queue


class RingBuffer(object):
    """Ring buffer to hold audio from audio stream"""

    def __init__(self, size=4096):
        self._buf = collections.deque(maxlen=size)

    def get_buf(self):
        return self._buf

    def lenght(self):
        return len(self._buf)

    def extend(self, data):
        """Adds data to the end of buffer"""
        self._buf.extend(data)

    def get(self, is_clear=True):
        """Retrieves data from the beginning of buffer and clears it"""
        tmp = bytes(bytearray(self._buf))
        is_clear and self._buf.clear()
        return tmp


class AudioBufferManager:
    """
    Manages an audio buffer, allowing addition and retrieval of audio data.
    """

    def __init__(self, audio_buffer: queue.Queue):
        """
        Args:
            audio_buffer (queue.Queue): Queue to be used as the audio buffer.
        """
        self.audio_buffer = audio_buffer
        self.total_samples = 0

    def add_to_buffer(self, audio_data):
        """
        Adds audio data to the buffer.

        Args:
            audio_data: Audio data to be added.
        """
        self.audio_buffer.put(audio_data)
        self.total_samples += len(audio_data) // 2

    def empty(self):
        return self.audio_buffer.empty()

    def clear_buffer(self):
        """Clears all audio data from the buffer."""
        while not self.audio_buffer.empty():
            try:
                self.audio_buffer.get_nowait()
            except queue.Empty:
                continue
        self.total_samples = 0

    def get_from_buffer(self, timeout: float = 0.05):
        """
        Retrieves audio data from the buffer.

        Args:
            timeout (float): Time (in seconds) to wait
              before raising a queue.Empty exception.

        Returns:
            The audio data chunk or None if the buffer is empty.
        """
        try:
            chunk = self.audio_buffer.get(timeout=timeout)
            self.total_samples -= len(chunk) // 2
            return chunk
        except queue.Empty:
            return None

    def get_buffered_seconds(self, rate: int) -> float:
        """
        Calculates the duration (in seconds) of the buffered audio data.

        Args:
            rate (int): Sample rate of the audio data.

        Returns:
            float: Duration of buffered audio in seconds.
        """
        return self.total_samples / rate



import logging
import asyncio
import uuid

from apipeline.frames.control_frames import EndFrame

from achatbot.common.types import DailyRoomBotArgs
from achatbot.common.interface import IBot
from achatbot.common.register import Register
from achatbot.common.session import Session
from achatbot.common.types import SessionCtx

register_daily_room_bots = Register('daily-room-bots')


class DailyRoomBot(IBot):

    def __init__(self, **args) -> None:
        self.args = DailyRoomBotArgs(**args)
        if self.args.bot_name is None or len(self.args.bot_name) == 0:
            self.args.bot_name = self.__class__.__name__

        self.task = None
        self._bot_config = self.args.bot_config
        self.session = Session(**SessionCtx(uuid.uuid4()).__dict__)

    def bot_config(self):
        return self._bot_config

    def run(self):
        try:
            asyncio.run(self.arun())
        except KeyboardInterrupt:
            logging.warning("Ctrl-C detected. Exiting!")

    async def arun(self):
        pass

    async def on_first_participant_joined(self, transport, participant):
        self.session.set_client_id(participant['id'])
        logging.info(f"First participant {participant['id']} joined")

    async def on_participant_left(self, transport, participant, reason):
        if self.task is not None:
            await self.task.queue_frame(EndFrame())
        logging.info("Partcipant left. Exiting.")

    async def on_call_state_updated(self, transport, state):
        logging.info("Call state %s " % state)
        if state == "left" and self.task is not None:
            await self.task.queue_frame(EndFrame())



# from distutils.core import setup
# from Cython.Build import cythonize
# import numpy

# setup(name='monotonic_align',
#       ext_modules=cythonize("core.pyx"),
#       include_dirs=[numpy.get_include()])



import logging
import os

from achatbot.common.types import MODELS_DIR, RECORDS_DIR, CosyVoiceTTSArgs
from achatbot.common import interface
from achatbot.common.factory import EngineClass, EngineFactory

from dotenv import load_dotenv
load_dotenv(override=True)


class TTSEnvInit():

    @staticmethod
    def getEngine(tag, **kwargs) -> interface.ITts | EngineClass:
        if "tts_coqui" in tag:
            from . import coqui_tts
        elif "tts_chat" in tag:
            from . import chat_tts
        elif "tts_pyttsx3" in tag:
            from . import pyttsx3_tts
        elif "tts_g" in tag:
            from . import g_tts
        elif "tts_edge" in tag:
            from . import edge_tts
        elif "tts_cosy_voice" in tag:
            from . import cosy_voice_tts
        # elif "tts_openai" in tag:
            # from . import openai_tts

        engine = EngineFactory.get_engine_by_tag(EngineClass, tag, **kwargs)
        return engine

    @staticmethod
    def initTTSEngine() -> interface.ITts | EngineClass:
        # tts
        tag = os.getenv('TTS_TAG', "tts_edge")
        kwargs = TTSEnvInit.map_config_func[tag]()
        engine = TTSEnvInit.getEngine(tag, **kwargs)
        logging.info(f"initTTSEngine: {tag}, {engine}")
        return engine

    @staticmethod
    def get_tts_chat_args() -> dict:
        kwargs = {}
        kwargs["local_path"] = os.getenv('LOCAL_PATH', os.path.join(
            MODELS_DIR, "2Noise/ChatTTS"))
        kwargs["source"] = os.getenv('TTS_CHAT_SOURCE', "custom")
        return kwargs

    @staticmethod
    def get_tts_coqui_args() -> dict:
        kwargs = {}
        kwargs["model_path"] = os.getenv('TTS_MODEL_PATH', os.path.join(
            MODELS_DIR, "coqui/XTTS-v2"))
        kwargs["conf_file"] = os.getenv(
            'TTS_CONF_FILE', os.path.join(MODELS_DIR, "coqui/XTTS-v2/config.json"))
        kwargs["reference_audio_path"] = os.getenv('TTS_REFERENCE_AUDIO_PATH', os.path.join(
            RECORDS_DIR, "me.wav"))
        kwargs["tts_stream"] = bool(os.getenv('TTS_STREAM', ""))
        return kwargs

    @staticmethod
    def get_tts_cosy_voice_args() -> dict:
        kwargs = CosyVoiceTTSArgs().__dict__
        return kwargs

    @staticmethod
    def get_tts_edge_args() -> dict:
        kwargs = {}
        kwargs["voice_name"] = os.getenv('TTS_VOICE', "zh-CN-XiaoxiaoNeural")
        kwargs["language"] = os.getenv('TTS_LANG', "zh")
        return kwargs

    @staticmethod
    def get_tts_g_args() -> dict:
        kwargs = {}
        kwargs["language"] = os.getenv('TTS_LANG', "zh")
        kwargs["speed_increase"] = float(os.getenv('SPEED_INCREASE', "1.5"))
        return kwargs

    # TAG : config
    map_config_func = {
        'tts_coqui': get_tts_coqui_args,
        'tts_cosy_voice': get_tts_cosy_voice_args,
        'tts_chat': get_tts_chat_args,
        'tts_edge': get_tts_edge_args,
        'tts_g': get_tts_g_args,
    }



from typing import AsyncGenerator
import logging
import asyncio
import struct
import time


from achatbot.common.interface import IRecorder
from achatbot.common.audio_stream.helper import RingBuffer
from achatbot.common.session import Session
from achatbot.common.types import SILENCE_THRESHOLD, RATE
from .base import AudioRecorder


class RMSRecorder(AudioRecorder, IRecorder):
    TAG = "rms_recorder"

    def __init__(self, **args) -> None:
        super().__init__(**args)

    def compute_rms(self, data):
        # Assuming data is in 16-bit samples
        format = "<{}h".format(len(data) // 2)
        ints = struct.unpack(format, data)

        # Calculate RMS
        sum_squares = sum(i ** 2 for i in ints)
        rms = (sum_squares / len(ints)) ** 0.5
        return rms

    async def record_audio(self, session: Session) -> list[bytes | None]:
        frames = []
        async for item in self._record_audio_generator(session):
            item and frames.append(item)

        return frames

    async def record_audio_generator(self, session: Session) -> AsyncGenerator[bytes, None]:
        async for item in self._record_audio_generator(session):
            yield item

    async def _record_audio_generator(self, session: Session) -> AsyncGenerator[bytes, None]:

        if self.stream_info.in_sample_rate != RATE:
            raise Exception(
                f"rms recorder's sampling rate of the audio just support 16000Hz at now")
        silent_chunks = 0
        audio_started = False
        silence_timeout = 0
        if "silence_timeout_s" in session.ctx.state:
            logging.info(
                f"rms recording with silence_timeout {session.ctx.state['silence_timeout_s']} s")
            silence_timeout = int(session.ctx.state['silence_timeout_s'])

        self.audio.start_stream()
        logging.debug(f"start {self.TAG} recording")
        start_time = time.time()
        if self.args.is_stream_callback is False:
            self.set_args(num_frames=self.stream_info.in_frames_per_buffer)
        while True:
            data = self.get_record_buf()
            if len(data) == 0:
                await asyncio.sleep(self.args.no_stream_sleep_time_s)
                continue
            rms = self.compute_rms(data)
            if audio_started:
                yield data
                if rms < SILENCE_THRESHOLD:
                    silent_chunks += 1
                    if silent_chunks > self.args.silent_chunks:
                        yield None
                        break
                else:
                    silent_chunks = 0
            elif rms >= SILENCE_THRESHOLD:
                yield data
                audio_started = True
            else:
                if silence_timeout > 0 \
                        and time.time() - start_time > silence_timeout:
                    logging.warning(f"rms recording silence timeout")
                    break

        self.audio.stop_stream()
        logging.debug(f"end {self.TAG} recording")


class WakeWordsRMSRecorder(RMSRecorder):
    TAG = "wakeword_rms_recorder"

    def __init__(self, **args) -> None:
        super().__init__(**args)

    async def _record_audio(self, session: Session) -> list[bytes]:
        sample_rate, frame_length = session.ctx.waker.get_sample_info()
        self.sample_rate, self.frame_length = sample_rate, frame_length
        self.set_args(frames_per_buffer=self.frame_length)

        # ring buffer
        pre_recording_buffer_duration = 3.0
        maxlen = int((sample_rate // frame_length) *
                     pre_recording_buffer_duration)
        self.audio_buffer = RingBuffer(maxlen)

        self.audio.start_stream()
        logging.info(
            f"start {self.TAG} recording; audio sample_rate: {self.sample_rate},frame_length:{self.frame_length}, audio buffer maxlen: {maxlen}")

        if self.args.is_stream_callback is False:
            self.set_args(num_frames=self.frame_length)

        while True:
            data = self.get_record_buf()
            if len(data) == 0:
                await asyncio.sleep(self.args.no_stream_sleep_time_s)
                continue
            session.ctx.read_audio_frames = data
            session.ctx.waker.set_audio_data(self.audio_buffer.get_buf())
            res = await session.ctx.waker.detect(session)
            if res is True:
                break
            self.audio_buffer.extend(data)

        self.audio.stop_stream()
        logging.debug("end wake words detector rms recording")

        if self.args.silence_timeout_s > 0:
            session.ctx.state["silence_timeout_s"] = self.args.silence_timeout_s

    async def record_audio(self, session: Session) -> list[bytes | None]:
        if session.ctx.waker is None:
            logging.warning(
                f"{self.TAG} no waker instance in session ctx, use {super().TAG}")
            return await super().record_audio(session)

        await self._record_audio(session)
        return await super().record_audio(session)

    async def record_audio_generator(self, session: Session) -> AsyncGenerator[bytes, None]:
        if session.ctx.waker is None:
            logging.warning(
                f"{self.TAG} no waker instance in session ctx, use {super().TAG}")
            async for item in super().record_audio_generator(session):
                yield item

        await self._record_audio(session)
        async for item in super().record_audio_generator(session):
            yield item



import logging

from .base import BaseVADAnalyzer
from achatbot.common.types import DAILY_WEBRTC_VAD_RESET_PERIOD_MS


class DailyWebRTCVADAnalyzer(BaseVADAnalyzer):
    TAG = "daily_webrtc_vad_analyzer"

    def __init__(self, **args):
        super().__init__(**args)
        from daily import Daily
        self._webrtc_vad = Daily.create_native_vad(
            reset_period_ms=DAILY_WEBRTC_VAD_RESET_PERIOD_MS,
            sample_rate=self._args.sample_rate,
            channels=self._args.num_channels
        )
        logging.debug("Loaded native WebRTC VAD")

    def voice_confidence(self, buffer) -> float:
        confidence = 0
        if len(buffer) > 0:
            confidence = self._webrtc_vad.analyze_frames(buffer)
        return confidence



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import tts_pb2 as tts__pb2

GRPC_GENERATED_VERSION = '1.65.0'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.65.0'
SCHEDULED_RELEASE_DATE = 'June 25, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in tts_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class TTSStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.LoadModel = channel.unary_unary(
            '/chat_bot.tts.TTS/LoadModel',
            request_serializer=tts__pb2.LoadModelRequest.SerializeToString,
            response_deserializer=tts__pb2.LoadModelResponse.FromString,
            _registered_method=True)
        self.SynthesizeUS = channel.unary_stream(
            '/chat_bot.tts.TTS/SynthesizeUS',
            request_serializer=tts__pb2.SynthesizeRequest.SerializeToString,
            response_deserializer=tts__pb2.SynthesizeResponse.FromString,
            _registered_method=True)


class TTSServicer(object):
    """Missing associated documentation comment in .proto file."""

    def LoadModel(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SynthesizeUS(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TTSServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'LoadModel': grpc.unary_unary_rpc_method_handler(
            servicer.LoadModel,
            request_deserializer=tts__pb2.LoadModelRequest.FromString,
            response_serializer=tts__pb2.LoadModelResponse.SerializeToString,
        ),
        'SynthesizeUS': grpc.unary_stream_rpc_method_handler(
            servicer.SynthesizeUS,
            request_deserializer=tts__pb2.SynthesizeRequest.FromString,
            response_serializer=tts__pb2.SynthesizeResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        'chat_bot.tts.TTS', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('chat_bot.tts.TTS', rpc_method_handlers)

 # This class is part of an EXPERIMENTAL API.


class TTS(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def LoadModel(request,
                  target,
                  options=(),
                  channel_credentials=None,
                  call_credentials=None,
                  insecure=False,
                  compression=None,
                  wait_for_ready=None,
                  timeout=None,
                  metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/chat_bot.tts.TTS/LoadModel',
            tts__pb2.LoadModelRequest.SerializeToString,
            tts__pb2.LoadModelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SynthesizeUS(request,
                     target,
                     options=(),
                     channel_credentials=None,
                     call_credentials=None,
                     insecure=False,
                     compression=None,
                     wait_for_ready=None,
                     timeout=None,
                     metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/chat_bot.tts.TTS/SynthesizeUS',
            tts__pb2.SynthesizeRequest.SerializeToString,
            tts__pb2.SynthesizeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



import logging
import math

from achatbot.common.types import (
    VAD_CHECK_PER_FRAMES,
    VAD_CHECK_ALL_FRAMES,
    WebRTCVADArgs,
)
from achatbot.common.session import Session
from .base import BaseVAD


class WebrtcVAD(BaseVAD):
    TAG = "webrtc_vad"

    def __init__(self, **args) -> None:
        import webrtcvad
        self.args = WebRTCVADArgs(**args)
        self.model = webrtcvad.Vad(self.args.aggressiveness)
        self.audio_buffer = None

    async def detect(self, session: Session):
        if self.args.check_frames_mode not in [VAD_CHECK_ALL_FRAMES, VAD_CHECK_PER_FRAMES]:
            return False

        # Number of audio frames per millisecond
        frame_length = int(self.args.sample_rate * self.args.frame_duration_ms / 1000)
        # just process channels:1 sample_width:2 audio
        num_frames = math.ceil(len(self.audio_buffer) / (2 * frame_length))
        speech_frames = 0
        logging.debug(
            f"{self.TAG} Speech detected audio_len:{len(self.audio_buffer)} frame_len:{frame_length} num_frames {num_frames}")

        for i in range(num_frames):
            start_byte = i * frame_length * 2
            end_byte = start_byte + frame_length * 2
            frame = self.audio_buffer[start_byte:end_byte]
            # print(len(frame))
            if self.model.is_speech(frame, self.args.sample_rate):
                speech_frames += 1
                if self.args.check_frames_mode == VAD_CHECK_PER_FRAMES:
                    logging.debug(f"{self.TAG} Speech detected in frame {i + 1}"
                                  f" of {num_frames}")
                    # await save_audio_to_file(frame, "webrtc_vad_frame.wav")
                    return True
        if self.args.check_frames_mode == VAD_CHECK_ALL_FRAMES:
            if speech_frames == num_frames:
                logging.debug(f"{self.TAG} Speech detected in {speech_frames} of "
                              f"{num_frames} frames")
            else:
                logging.debug(f"{self.TAG} Speech not detected in all {num_frames} frames")
            return speech_frames == num_frames

        logging.debug(f"{self.TAG} Speech not detected in any of {num_frames} frames")
        return False



import threading
import traceback
import asyncio
import logging
import time
import sys
import os

import uuid

from achatbot.common import interface
from achatbot.common.session import Session
from achatbot.common.utils.wav import save_audio_to_file
from achatbot.common.types import SessionCtx, RECORDS_DIR
from achatbot.common.utils.time import get_current_formatted_time
if os.getenv("INIT_TYPE", 'env') == 'yaml_config':
    from achatbot.cmd.init import YamlConfig as init
else:
    from achatbot.cmd.init import Env as init


class TerminalChatClient:
    def __init__(self, is_save_record=False, is_save_play_chunks=False) -> None:
        self.sid = str(uuid.uuid4())
        self.session = Session(**SessionCtx(self.sid).__dict__)

        audio_out_stream = init.initAudioOutStreamEngine()
        self.player = init.initPlayerEngine()
        self.player.set_out_stream(audio_out_stream)

        audio_in_stream = init.initAudioInStreamEngine()
        self.recorder = init.initRecorderEngine()
        self.recorder.set_in_stream(audio_in_stream)

        self.waker = init.initWakerEngine()
        self.vad = init.initVADEngine()

        self.start_record_event = threading.Event()
        self.play_chunks = []
        self.is_save_play_chunks = is_save_play_chunks
        self.is_save_record = is_save_record

    def run(self, conn: interface.IConnector):
        if self.is_save_play_chunks:
            self.player.set_args(on_play_chunk=self.on_play_chunk,
                                 on_play_end=self.on_play_end)
        self.player.open()
        self.player.start(self.session)
        play_t = threading.Thread(target=self.loop_recv,
                                  args=(conn,))
        play_t.start()

        self.recorder.open()
        record_t = threading.Thread(target=self.loop_record,
                                    args=(conn,))
        record_t.start()

        record_t.join()
        play_t.join()
        self.player.close()
        self.recorder.close()

    def on_wakeword_detected(self, session: Session, data):
        if "bot_name" in session.ctx.state:
            print(f"{session.ctx.state['bot_name']}~ ",
                  end="", flush=True, file=sys.stderr)

    def on_play_end(self, session: Session):
        sample_info = init.get_stream_info()
        # @todo: need async save i/o
        asyncio.run(save_audio_to_file(
            b''.join(self.play_chunks),
            self.session.get_paly_audio_name(),
            sample_rate=sample_info["rate"],
            channles=sample_info["channels"],
            sample_width=sample_info["sample_width"],
            audio_dir=RECORDS_DIR))
        logging.info(
            f"play end with session.ctx {session.ctx}, saved chunks_len {len(self.play_chunks)}")

    def on_play_chunk(self, session: Session, sub_chunk: bytes):
        logging.info(
            f"play chunk with session.ctx {session.ctx}, sub_chunk_len {len(sub_chunk)}")
        self.play_chunks.append(sub_chunk)

    def loop_record(self, conn: interface.IConnector):
        if self.waker is not None:
            self.waker.set_args(on_wakeword_detected=self.on_wakeword_detected)
        logging.info(
            f"loop_record starting with session ctx: {self.session.ctx}")
        print(
            f"start loop_record with {self.recorder.TAG} ...", flush=True, file=sys.stderr)
        while True:
            try:
                print(f"-- chat round {self.session.chat_round} --",
                      flush=True, file=sys.stdout)
                # self.start_record_event.clear()
                print(f"\n({get_current_formatted_time()}) me >> ",
                      end="", flush=True, file=sys.stderr)

                logging.info(f"start record audio")
                self.session.ctx.waker = self.waker
                self.session.ctx.vad = self.vad
                frames = asyncio.run(self.recorder.record_audio(self.session))
                if len(frames) == 0:
                    logging.info(f"record_audio return empty, continue")
                    continue
                data = b''.join(frames)
                conn.send(("RECORD_FRAMES", data, self.session), 'fe')
                if self.is_save_record:
                    asyncio.run(save_audio_to_file(
                        data, self.session.get_record_audio_name(),
                        audio_dir=RECORDS_DIR))
                self.session.increment_chat_round()

                if self.start_record_event.is_set():
                    logging.info(f"start record audio event is set, clear it!")
                    self.start_record_event.clear()
                logging.info(f"wait start record audio event")
                self.start_record_event.wait()
            except Exception as ex:
                ex_trace = traceback.format_exc()
                logging.warning(
                    f"loop_record Exception {ex} sid:{self.session.ctx.client_id} Trace {ex_trace}")
                time.sleep(1)

    def loop_recv(self, conn: interface.IConnector):
        print(
            f"start loop_recv with {self.player.TAG} ...", flush=True, file=sys.stderr)
        llm_gen_segments = 0
        while True:
            try:
                res = conn.recv('fe')
                if res is None:
                    continue

                msg, recv_data, session = res
                if msg is None or msg.lower() == "stop":
                    break
                if isinstance(recv_data, str):
                    logging.info(
                        f'FE Received: {msg} recv_data: {recv_data}, session: {session}')
                else:
                    logging.info(
                        f'FE Received: {msg} len(recv_data): {len(recv_data)}, session: {session}')

                if msg == "BE_EXCEPTION":
                    print(f"\nBE exception: {recv_data.strip()}",
                          end="", flush=True, file=sys.stderr)
                    self.start_record_event.set()
                    llm_gen_segments = 0
                    continue

                if session.ctx.client_id != self.session.ctx.client_id:
                    logging.warning(
                        f"session.ctx.client_id: {session.ctx.client_id} != self.session.ctx.client_id: {self.session.ctx.client_id}")
                    continue

                if msg == "PLAY_FRAMES":
                    self.session.ctx.state["tts_chunk"] = recv_data
                    self.player.play_audio(self.session)
                elif msg == "PLAY_FRAMES_DONE":
                    self.player.stop(self.session)
                    self.start_record_event.set()
                    llm_gen_segments = 0
                    self.player.start(self.session)
                elif msg == "LLM_GENERATE_TEXT":
                    if llm_gen_segments == 0:
                        bot_name = self.session.ctx.state["bot_name"] if "bot_name" in self.session.ctx.state else "bot"
                        logging.info(f"bot_name: {bot_name}")
                        print(f"\n({get_current_formatted_time()}) {bot_name} >> ",
                              end="", flush=True, file=sys.stderr)
                    print(recv_data.strip(), end="",
                          flush=True, file=sys.stderr)
                    llm_gen_segments += 1
                elif msg == "LLM_GENERATE_DONE":
                    print("\n", end="", flush=True, file=sys.stderr)
                    llm_gen_segments = 0
                elif msg == "ASR_TEXT":
                    print(recv_data.strip(), end="",
                          flush=True, file=sys.stderr)
                elif msg == "ASR_TEXT_DONE":
                    print("\n", end="", flush=True, file=sys.stderr)
                else:
                    logging.warning(f"unsupport msg {msg}")
            except Exception as ex:
                ex_trace = traceback.format_exc()
                logging.warning(f"loop_recv Exception {ex}, trace: {ex_trace}")
                self.start_record_event.set()
                llm_gen_segments = 0

    @staticmethod
    def clear_console():
        os.system('clear' if os.name == 'posix' else 'cls')



from dataclasses import dataclass


from apipeline.frames.sys_frames import SystemFrame


@dataclass
class BotInterruptionFrame(SystemFrame):
    """Emitted by when the bot should be interrupted. This will mainly cause the
    same actions as if the user interrupted except that the
    UserStartedSpeakingFrame and UserStoppedSpeakingFrame won't be generated.

    """
    pass
